@@ -94,7 +94,6 @@ let%test_unit "roundtrip" =
 external is_mmapped : t_frozen -> bool = "bigstring_is_mmapped_stub" [@@noalloc]
 
 let%test "bigstring created with create are not mmapped" = not (is_mmapped (create 2))
-
 let init = init
 
 let try_setters z =
@@ -149,6 +148,15 @@ let unsafe_set_uint64_be = unsafe_set_uint64_be
 let unsafe_set_int32_le = unsafe_set_int32_le
 let unsafe_set_int32_be = unsafe_set_int32_be
 
+module Local = struct
+  open Local
+
+  let get_int64_t_le = get_int64_t_le
+  let get_int64_t_be = get_int64_t_be
+  let unsafe_get_int64_t_le = unsafe_get_int64_t_le
+  let unsafe_get_int64_t_be = unsafe_get_int64_t_be
+end
+
 let%test_module "truncating setters (should end in [_trunc] or begin with [unsafe_])" =
   (module struct
     let test setters z =
@@ -159,11 +167,15 @@ let%test_module "truncating setters (should end in [_trunc] or begin with [unsaf
     ;;
 
     let%expect_test "all word sizes" =
-      test [ unsafe_set_int8 ] 0x9080;
+      test [ (fun buf ~pos value -> unsafe_set_int8 buf ~pos value) ] 0x9080;
       [%expect {| 80 0 0 0 0 0 0 0 ; |}];
-      test [ unsafe_set_uint8 ] (-1);
+      test [ (fun buf ~pos value -> unsafe_set_uint8 buf ~pos value) ] (-1);
       [%expect {| ff 0 0 0 0 0 0 0 ; |}];
-      test [ unsafe_set_int16_le; unsafe_set_int16_be ] 0x90_8070;
+      test
+        [ (fun buf ~pos value -> unsafe_set_int16_le buf ~pos value)
+        ; (fun buf ~pos value -> unsafe_set_int16_be buf ~pos value)
+        ]
+        0x90_8070;
       [%expect {| 70 80 0 0 0 0 0 0 ; 80 70 0 0 0 0 0 0 ; |}];
       test [ unsafe_set_uint16_le; unsafe_set_uint16_be ] (-1);
       [%expect {| ff ff 0 0 0 0 0 0 ; ff ff 0 0 0 0 0 0 ; |}];
@@ -175,7 +187,11 @@ let%test_module "truncating setters (should end in [_trunc] or begin with [unsaf
 
     let%expect_test (_ [@tags "64-bits-only"]) =
       Option.iter (Int64.to_int 0x90_8070_6050L) ~f:(fun z ->
-        test [ unsafe_set_int32_le; unsafe_set_int32_be ] z);
+        test
+          [ (fun buf ~pos value -> unsafe_set_int32_le buf ~pos value)
+          ; (fun buf ~pos value -> unsafe_set_int32_be buf ~pos value)
+          ]
+          z);
       [%expect {| 50 60 70 80 0 0 0 0 ; 80 70 60 50 0 0 0 0 ; |}]
     ;;
   end)
@@ -192,8 +208,7 @@ let%test_module "truncating getters (should end in [_trunc] or begin with [unsaf
     let test getter =
       List.iter
         [ 0x81 (* positive if top bit truncated *)
-        ; 0xc1
-          (* negative if top bit truncated *)
+        ; 0xc1 (* negative if top bit truncated *)
         ]
         ~f:(fun first_byte ->
           let i = getter (getter_t ~first_byte) ~pos:0 in
@@ -334,7 +349,11 @@ let set_int64_t_le = set_int64_t_le
 let%expect_test "basic int64 setters" =
   try_setters
     0x0102030405060708L
-    [ unsafe_set_int64_t_be; unsafe_set_int64_t_le; set_int64_t_be; set_int64_t_le ]
+    [ (fun t ~pos x -> unsafe_set_int64_t_be t ~pos x)
+    ; (fun t ~pos x -> unsafe_set_int64_t_le t ~pos x)
+    ; (fun t ~pos x -> set_int64_t_be t ~pos x)
+    ; (fun t ~pos x -> set_int64_t_le t ~pos x)
+    ]
   |> printf !"%{sexp#hum:int list Or_error.t list}\n";
   [%expect
     {|
@@ -350,12 +369,21 @@ let get_int64_t_le = get_int64_t_le
 let%expect_test "basic int64 getters" =
   try_getters
     ~first_bigstring_byte:1
-    [ unsafe_get_int64_t_be; unsafe_get_int64_t_le; get_int64_t_be; get_int64_t_le ]
+    [ unsafe_get_int64_t_be
+    ; unsafe_get_int64_t_le
+    ; get_int64_t_be
+    ; (fun t ~pos -> Int64.( + ) 0L (get_int64_t_le t ~pos))
+    ; (fun t ~pos -> (Local.unsafe_get_int64_t_be t ~pos |> globalize_int64) [@nontail])
+    ; (fun t ~pos -> (Local.unsafe_get_int64_t_le t ~pos |> globalize_int64) [@nontail])
+    ; (fun t ~pos -> (Local.get_int64_t_be t ~pos |> globalize_int64) [@nontail])
+    ; (fun t ~pos -> (Local.get_int64_t_le t ~pos |> globalize_int64) [@nontail])
+    ]
   |> printf !"%{sexp#hum:Int64.Hex.t Or_error.t list}\n";
   [%expect
     {|
       ((Ok 0x102030405060708) (Ok 0x807060504030201) (Ok 0x102030405060708)
-       (Ok 0x807060504030201)) |}]
+       (Ok 0x807060504030201) (Ok 0x102030405060708) (Ok 0x807060504030201)
+       (Ok 0x102030405060708) (Ok 0x807060504030201)) |}]
 ;;
 
 let unsafe_set_int32_t_be = unsafe_set_int32_t_be
@@ -486,6 +514,76 @@ let%expect_test "basic find" =
   [%expect {| (Invalid_argument "find: len < 0") |}]
 ;;
 
+external unsafe_memmem
+  :  haystack:t
+  -> needle:t
+  -> haystack_pos:int
+  -> haystack_len:int
+  -> needle_pos:int
+  -> needle_len:int
+  -> int
+  = "bigstring_memmem_bytecode" "bigstring_memmem"
+[@@noalloc]
+
+let%expect_test "basic unsafe_memmem" =
+  let haystack = "foo bar baz qwux" |> of_string in
+  let t ~haystack_pos ~haystack_len ~needle_pos ~needle_len needle =
+    let result =
+      unsafe_memmem
+        ~haystack
+        ~needle:(of_string needle)
+        ~haystack_pos
+        ~haystack_len
+        ~needle_pos
+        ~needle_len
+    in
+    print_s [%sexp (result : int)]
+  in
+  t ~haystack_pos:0 ~haystack_len:16 ~needle_pos:0 ~needle_len:4 "nope";
+  [%expect {| -1 |}];
+  t ~haystack_pos:0 ~haystack_len:16 ~needle_pos:0 ~needle_len:4 "qwux";
+  [%expect {| 12 |}];
+  t ~haystack_pos:0 ~haystack_len:15 ~needle_pos:0 ~needle_len:4 "qwux";
+  [%expect {| -1 |}];
+  t ~haystack_pos:0 ~haystack_len:16 ~needle_pos:1 ~needle_len:3 "ZfooZ";
+  [%expect {| 0 |}];
+  t ~haystack_pos:1 ~haystack_len:16 ~needle_pos:1 ~needle_len:3 "ZfooZ";
+  [%expect {| -1 |}];
+  t ~haystack_pos:1 ~haystack_len:16 ~needle_pos:1 ~needle_len:3 "Zoo Z";
+  [%expect {| 1 |}]
+;;
+
+let memmem = memmem
+
+let%expect_test "basic memmem" =
+  let haystack = "foo bar baz qwux" |> of_string in
+  let t ?haystack_pos ?haystack_len ?needle_pos ?needle_len needle =
+    let result =
+      memmem
+        ~haystack
+        ~needle:(of_string needle)
+        ?haystack_pos
+        ?haystack_len
+        ?needle_pos
+        ?needle_len
+        ()
+    in
+    print_s [%sexp (result : int option)]
+  in
+  t "nope";
+  [%expect {| () |}];
+  t "qwux";
+  [%expect {| (12) |}];
+  t ~haystack_len:15 "qwux";
+  [%expect {| () |}];
+  t ~needle_pos:1 ~needle_len:3 "ZfooZ";
+  [%expect {| (0) |}];
+  t ~haystack_pos:1 ~needle_pos:1 ~needle_len:3 "ZfooZ";
+  [%expect {| () |}];
+  t ~haystack_pos:1 ~needle_pos:1 ~needle_len:3 "Zoo Z";
+  [%expect {| (1) |}]
+;;
+
 let get_opt_len = get_opt_len
 
 let%expect_test "basic get_opt_len" =
@@ -535,6 +633,19 @@ let%expect_test "basic char setters" =
        (Ok (120 0 0 0 0 0 0 0))) |}]
 ;;
 
+external unsafe_set : t_frozen -> int -> char -> unit = "%caml_ba_unsafe_set_1"
+
+let%expect_test "basic char unsafe setters" =
+  try_setters
+    'x'
+    [ memset ~len:0; memset ~len:1; memset ~len:2; (fun t ~pos -> set t pos) ]
+  |> printf !"%{sexp#hum:int list Or_error.t list}\n";
+  [%expect
+    {|
+      ((Ok (0 0 0 0 0 0 0 0)) (Ok (120 0 0 0 0 0 0 0)) (Ok (120 120 0 0 0 0 0 0))
+       (Ok (120 0 0 0 0 0 0 0))) |}]
+;;
+
 external get : t_frozen -> int -> char = "%caml_ba_ref_1"
 
 let%expect_test "basic char getters" =
@@ -543,6 +654,14 @@ let%expect_test "basic char getters" =
   [%expect {| ((Ok "\001")) |}]
 ;;
 
+external unsafe_get : t -> int -> char = "%caml_ba_unsafe_ref_1"
+
+let%expect_test "basic unsafe char getters" =
+  try_getters ~first_bigstring_byte:1 [ (fun t ~pos -> unsafe_get t pos) ]
+  |> printf !"%{sexp#hum:char Or_error.t list}\n";
+  [%expect {| ((Ok "\001")) |}]
+;;
+
 let check_args = check_args
 
 let%expect_test "basic check_args" =
@@ -612,6 +731,18 @@ let%expect_test "basic equal" =
     List.iter strings ~f:test)
 ;;
 
+let%expect_test ("local allocation does not heap allocate" [@tags "64-bits-only"]) =
+  let t = init 8 ~f:Char.of_int_exn in
+  Expect_test_helpers_core.require_no_allocation [%here] (fun () ->
+    let x = Local.get_int64_t_be t ~pos:0 in
+    set_int64_t_be t ~pos:0 x [@nontail]);
+  [%expect ""];
+  Expect_test_helpers_core.require_no_allocation [%here] (fun () ->
+    let x = Local.get_int64_t_le t ~pos:0 in
+    set_int64_t_le t ~pos:0 x [@nontail]);
+  [%expect ""]
+;;
+
 type nonrec t = t
 type nonrec t_frozen = t_frozen [@@deriving compare, hash, sexp]
 
